/* Copyright 2015-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

#include "qubx_idealize_mdl.h"

#include <algorithm>
#include <iostream>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <vector>

#include "qub_mean.h"

using std::vector;
using std::cerr;
using std::cout;
using std::endl;
using std::max;
using std::min;
using std::sort;

float calc_rss(float *data, int n, float& mean)
{
  IncrementalRSS calc;
  calc.add_n(data, n);
  mean = (float) calc.mean;
  return (float) calc.rss;
}

float best_2split(float *data, int offset, int n, int Lmin, int& i_best)
// Finds the best split point within the interval data[offset..offset+n-1] and returns the rss1.
// out: i_best: the optimal right half begins at data[offset+i_best]
{
  IncrementalRSS L, R;
  double rss1, rss1_best = 0.0;
  data += offset;
  
  // initialize right stats with entire interval
  R.add_n(data, n);
  // transfer Lmin points to left stats
  for (int i=0; i<Lmin; ++i)
    R.remove(data[i]);
  L.add_n(data, Lmin);

  int Nsplit = n - 2*Lmin + 1;
  for (int i=0; i<Nsplit; ++i) {
    rss1 = L.rss + R.rss;
    if ( (i == 0) || (rss1 < rss1_best) ) {
      i_best = Lmin+i;
      rss1_best = rss1;
    }
    R.remove(data[i+Lmin]);
    L.add(data[i+Lmin]);
  }
  
  return (float) rss1_best;
}

float best_3split(float *data, int offset, int n, int Lmin, int& split1, int& split2)
// Finds the best two split points within the interval data[offset..offset+n-1] and returns the rss2.
// out: split1, split2: the optimal center segment is data[offset+split1..offset+split2-1]
{
  // We decompose this as
  //   argmin[split1, split2](RSS2(split1, split2))
  //        == argmin[split1](RSS(1..split1-1) + best_2split(data, offset+split1, n-split1, ...)
  // and compute each possible split1 in parallel
  vector<float> RSS2(n);
  vector<int> S2(n);
  int iLim = n-2*Lmin+1;
  data += offset;
#pragma omp parallel
  {
    IncrementalRSS A;
    int nA = 0;
    // schedule(static,1) assigns successive values of i to different processors, rather than assigning them in blocks.
    // It's important for load balancing because best_2split's n gets smaller with each i, so earlier iterations take longer.
#pragma omp for schedule(static,1)
    for (int i=Lmin; i<iLim; ++i) {
      if ( nA < i ) {
	A.add_n(data+nA, i-nA);
	nA = i;
      }
      int splitBC;
      float rssBC = best_2split(data, i, n-i, Lmin, splitBC);
      RSS2[i] = ((float) A.rss) + rssBC;
      S2[i] = splitBC+i;
    }
  }

  // non-parallel: find which split1 is best
  float rss2_best = RSS2[Lmin];
  split1 = Lmin;
  for (int i=Lmin+1; i<iLim; ++i) {
    if ( RSS2[i] < rss2_best ) {
      rss2_best = RSS2[i];
      split1 = i;
    }
  }
  split2 = S2[split1];
  return rss2_best;
}


class qubx_idealize_mdl_interval {
  // represents one dwell
public:
  int offset; // index in data[]
  int n; // number of datapoints
  float mean;
  float rss; // residual sum squares from observed mean
  void update(float *data) {
    // recalculates mean and rss
    rss = calc_rss(data+offset, n, mean);
  }
};

class qubx_idealize_mdl_intervals {
  // Manages a fixed store of interval records.
  // For N data points, there can be at most N intervals (actually, N/Lmin ?).
  // This container returns pointers to the records, which are enqueued and sorted variously in qubx_idealize_mdl_qs.
private:
  float *data;
  vector<qubx_idealize_mdl_interval> intervals;
  int next;
public:
  qubx_idealize_mdl_intervals(float *data_, int N)
    : data(data_), intervals(N), next(0)
  {}
  qubx_idealize_mdl_interval* make(int offset, int n) {
    // Returns a new interval
    qubx_idealize_mdl_interval *iv = & intervals[next++];
    iv->offset = offset;
    iv->n = n;
    return iv;
  }
  qubx_idealize_mdl_interval* split(qubx_idealize_mdl_interval *iv, int splitp) {
    // Modifies iv so its length == splitp, and returns a new interval containing the remainder of the original (starting at iv->offset+splitp).
    qubx_idealize_mdl_interval *iv2 = & intervals[next++];
    iv2->offset = iv->offset + splitp;
    iv2->n = iv->n - splitp;
    iv->n -= iv2->n;
    return iv2;
  }
};


// in C++11, these comparison functions can be replaced by in-line lambda expressions

bool qubx_idealize_mdl_cmp_offset(qubx_idealize_mdl_interval *a, qubx_idealize_mdl_interval *b) {
  // to sort intervals by offset increasing
  return a->offset < b->offset;
}

class qubx_idealize_mdl_cmp_ix_by_size {
  // to sort indices of dones by n decreasing
public:
  vector<qubx_idealize_mdl_interval*>& v;
  qubx_idealize_mdl_cmp_ix_by_size(vector<qubx_idealize_mdl_interval*>& dones)
    : v(dones)
  {}
  bool operator()(int a, int b) {
    return (v[a]->n > v[b]->n);
  }
};


class qubx_idealize_mdl_qs {
  // Manages lists of intervals which may be split, and a "done" list of intervals which can't be split further.
  // Starting with a single interval (0, N), we try recursively to split each interval, first in 2 pieces, then in 3, until
  // the un-split interval has a shorter description length than any possible split.
  // To do that, we keep lists of intervals to try splitting in 2 and 3 pieces respectively, and a list of intervals which have
  // failed to split either way, and are therefore done.
  // One pass consists of trying to split all the intervals from last_try2() into 2 pieces, then
  //     if the MDL is smaller: append the two half-intervals to the next try2 list
  //     else: append the interval to the next try3 list
  //     next: do the same with the try3 list;
  //         success: append the three sub-intervals to the next try2 list
  //         failure: append the interval to the done list
  // We repeat until all intervals are in the done list.
private:
  // Instead of swapping the last and next lists, we alternate "which" between 0 and 1;
  // processing items from [(which+1)%2], appending to [which]
  vector< vector<qubx_idealize_mdl_interval*> > tries2, tries3;
  int which;
  int Lmin;
  int n2, n3;
public:
  vector<qubx_idealize_mdl_interval*> dones;
  int Ntry2, Ntry3;

  qubx_idealize_mdl_qs(int Lmin_)
    : tries2(2), tries3(2), which(0), Lmin(Lmin_), n2(0), n3(0), Ntry2(0), Ntry3(0)
  {}
  void swap() {
    // Puts items accumulated by try2() and try3() into last_try2() and last_try3() respectively.
    //last_try2().clear();
    //last_try3().clear();
    Ntry2 = n2;
    Ntry3 = n3;
    n2 = n3 = 0;
    which = (which + 1) % 2;
  }
  void try2(qubx_idealize_mdl_interval *iv) // appends an interval to the next pass's 2 split trials
  { tries2[which].resize(max((int)tries2[which].size(), n2+1)); tries2[which][n2++] = iv; }
  void try3(qubx_idealize_mdl_interval *iv) // appends an interval to the next pass's 3 split trials
  { tries3[which].resize(max((int)tries3[which].size(), n3+1)); tries3[which][n3++] = iv; }
  void done(qubx_idealize_mdl_interval *iv) // appends an interval to the done list
  { dones.push_back(iv); }
  vector<qubx_idealize_mdl_interval*>& last_try2() // returns a reference to the last 2 split list (ready to process)
  { return tries2[(which+1)%2]; }
  vector<qubx_idealize_mdl_interval*>& last_try3() // returns a reference to the last 3 split list
  { return tries3[(which+1)%2]; }
  bool all_done() {
    // if last_try2() and last_try3() are empty, sorts dones by offset increasing and returns true.
    if ( Ntry2 || Ntry3 )
      return false;
    sort(dones.begin(), dones.end(), qubx_idealize_mdl_cmp_offset);
    return true;
  }

  void combine_close_levels(float *data, float min_delta) {
    // from longest to shortest, try to join each done event with its neighbors
    vector<int> ix_by_size(dones.size());
    int Nd = (int) dones.size();
    for (int i=Nd-1; i>=0; --i)
      ix_by_size[i] = i;
    sort(ix_by_size.begin(), ix_by_size.end(), qubx_idealize_mdl_cmp_ix_by_size(dones));
    for (int i=0; i<Nd; ++i) {
      qubx_idealize_mdl_interval *iv = dones[ix_by_size[i]];
      if ( iv->n ) {
	float lo = iv->mean - min_delta;
	float hi = iv->mean + min_delta;
	int ivn = iv->n;
	for (int j=ix_by_size[i]-1; j>=0; --j) {
	  qubx_idealize_mdl_interval *jv = dones[j];
	  if ( (jv->n == 0 ) || (jv->mean <= lo) || (jv->mean >= hi) )
	    break;
	  lo = max(lo, jv->mean - min_delta);
	  hi = min(hi, jv->mean + min_delta);
	  iv->n += jv->n;
	  jv->n = 0;
	  iv->offset = jv->offset;
	}
	for (int j=ix_by_size[i]+1; j<Nd; ++j) {
	  qubx_idealize_mdl_interval *jv = dones[j];
	  if ( (jv->n == 0 ) || (jv->mean <= lo) || (jv->mean >= hi) )
	    break;
	  lo = max(lo, jv->mean - min_delta);
	  hi = min(hi, jv->mean + min_delta);
	  iv->n += jv->n;
	  jv->n = 0;
	}
	if ( ivn != iv->n ) {
	  iv->update(data);
	}
      }
    }
    vector<qubx_idealize_mdl_interval*> last_dones(dones.size());
    std::copy(dones.begin(), dones.end(), last_dones.begin());
    int Ndone = 0;
    for (vector<qubx_idealize_mdl_interval*>::iterator di=last_dones.begin(); di!=last_dones.end(); ++di)
      if ( (*di)->n )
	dones[Ndone++] = *di;
    dones.resize(Ndone);
  }
};



void update_intervals(qubx_idealize_mdl_qs& qs, float *data, bool all_done) {
  // re-calculates mean and rss0 of all relevant intervals (if all_done: dones; else: last_try2,last_try3)
#pragma omp parallel
  {
    if ( all_done ) {
      int Niv = (int) qs.dones.size();
#pragma omp for schedule(guided)
      for (int i=0; i<Niv; ++i) {
	qs.dones[i]->update(data);
      }
    } else {
      int Niv = (int) qs.last_try2().size();
#pragma omp for schedule(guided)
      for (int i=0; i<Niv; ++i) {
	qs.last_try2()[i]->update(data);
      }
      Niv = (int) qs.last_try3().size();
#pragma omp for schedule(guided)
      for (int i=0; i<Niv; ++i) {
	qs.last_try3()[i]->update(data);
      }
    }
  }
}


void process_try2(qubx_idealize_mdl_qs& qs, qubx_idealize_mdl_intervals& ivs, float *data,
		  int Lmin, bool split3) {
  // each interval in last_try2(): evaluate all possible 2-splits.
  vector<int> splits(qs.Ntry2);
#pragma omp parallel for schedule(guided)
  for (int i=0; i<qs.Ntry2; ++i) {
    qubx_idealize_mdl_interval *iv = qs.last_try2()[i];
    double L0 = 0.5 * (log((double)iv->n) + ((double)iv->n)*log(iv->rss/(double)iv->n));
    float rss1 = best_2split(data, iv->offset, iv->n, Lmin, splits[i]);
    double L1 = 0.5 * (log((double)iv->n) + log((double)splits[i]) + log((double)(iv->n - splits[i])) + iv->n*log(rss1/(double)iv->n));
    if ( L1 >= L0 )
      splits[i] = 0; // better with no split
  }

  for (int i=0; i<qs.Ntry2; ++i) {
    qubx_idealize_mdl_interval *iv = qs.last_try2()[i];
    if ( splits[i] ) {
      // successful split: queue halves for re-splitting unless too short
      qubx_idealize_mdl_interval *iv2 = ivs.split(iv, splits[i]);
      if ( iv->n >= 2*Lmin )
	qs.try2(iv);
      else
	qs.done(iv);
      if ( iv2->n >= 2*Lmin )
	qs.try2(iv2);
      else
	qs.done(iv2);
    } else if ( split3 && (iv->n >= 3*Lmin) ) {
      qs.try3(iv);
    } else {
      qs.done(iv);
    }
  }
}

void process_try3(qubx_idealize_mdl_qs& qs, qubx_idealize_mdl_intervals& ivs, float *data,
		  int Lmin, bool& no_interrupt) {
  // last_try3() has intervals that failed to split into two pieces.
  for (int i=0; no_interrupt && i<qs.Ntry3; ++i) {
    qubx_idealize_mdl_interval *iv = qs.last_try3()[i];
    double L0 = 0.5 * (log((double)iv->n) + iv->n*log(iv->rss/(double)iv->n));
    int split, split2;
    float rss2 = best_3split(data, iv->offset, iv->n, Lmin, split, split2);
    double L2 = 0.5 * (2*log((double)iv->n) + log((double)split) + log((double)(split2-split)) + log((double)(iv->n-split2)) + iv->n*log(rss2/(double)iv->n));
    if ( L2 < L0 ) {
      //cout << "split3: " << iv->offset << " at " << split << " and " << split2 << " of " << iv->n << endl;
      qubx_idealize_mdl_interval *iv3 = ivs.split(iv, split2);
      if ( iv3->n >= 2*Lmin )
	qs.try2(iv3);
      else
	qs.done(iv3);
      qubx_idealize_mdl_interval *iv2 = ivs.split(iv, split);
      if ( iv2->n >= 2*Lmin )
	qs.try2(iv2);
      else
	qs.done(iv2);
      if ( iv->n >= 2*Lmin )
	qs.try2(iv);
      else
	qs.done(iv);
    } else {
      qs.done(iv);
    }
  }
}

// main work function
int qubx_idealize_mdl_run(float *data, int N, int Lmin, bool split3, bool combineClose, float combineDelta,
			  vector<int>& counts, vector<float>& amps, vector<float>& stds, bool& no_interrupt)
{  // returns dwell count
  qubx_idealize_mdl_intervals ivs(data, N); // interval storage
  qubx_idealize_mdl_qs qs(Lmin); // interval work/done queues
  // start with one interval
  qubx_idealize_mdl_interval *iv = ivs.make(0, N);
  // pass 0: enqueue for splitting if large enough
  if ( N >= 2*Lmin )
    qs.try2(iv);
  else if ( N > 0 )
    qs.done(iv);
  qs.swap(); // ready for next pass
  
  while ( no_interrupt && ! qs.all_done() ) {
    update_intervals(qs, data, false);
    process_try2(qs, ivs, data, Lmin, split3);

    if ( no_interrupt && qs.Ntry3 )
      process_try3(qs, ivs, data, Lmin, no_interrupt);

    qs.swap();
  }
  // all_done has sorted qs.dones by offset
  update_intervals(qs, data, true);

  if ( combineClose )
    qs.combine_close_levels(data, combineDelta);

  // copy them into the output arrays

  counts.clear();
  amps.clear();
  stds.clear();
  int Ndwell = 0;
  
  if ( no_interrupt ) {
    Ndwell = (int) qs.dones.size();
    counts.resize(Ndwell);
    amps.resize(Ndwell);
    stds.resize(Ndwell);
    for (int i=0; i<Ndwell; ++i) {
      qubx_idealize_mdl_interval *ivt = qs.dones[i];
      counts[i] = ivt->n;
      amps[i] = ivt->mean;
      stds[i] = (ivt->rss > 0.0) ? ((float) sqrt(ivt->rss / (double)ivt->n)) : 0.0f;
    }
  }
  
  return Ndwell;
}




// for the C-convention shared library interface:

class qubx_idealizer_mdl {
  // one idealization session, with opencl resources and storage for the outputs
private:
  vector<int> last_counts;
  vector<float> last_amps, last_stds;
  bool no_interrupt;
public:
  qubx_idealizer_mdl()
  {}
  ~qubx_idealizer_mdl()
  {}
  int idealize(float *data, int N, int Lmin, int split3, int combineClose, float combineDelta)
  {
    no_interrupt = true;
    return qubx_idealize_mdl_run(data, N, Lmin, (bool) split3, combineClose, combineDelta, last_counts, last_amps, last_stds, no_interrupt);
  }
  void interrupt() // stops idealize() from another thread
  {
    no_interrupt = false;
  }
  int get_dwells(int *counts, float *amps, float *stds, int capacity)
  {
    int N = min(capacity, (int) last_counts.size());
    memcpy(counts, &last_counts[0], N*sizeof(int));
    memcpy(amps, &last_amps[0], N*sizeof(float));
    if ( stds )
      memcpy(stds, &last_stds[0], N*sizeof(float));
    return N;
  }
};


class qubx_idealize_mdl_context {
  // global resources such as compiled OpenCL kernel
public:
  qubx_idealize_mdl_context()
  {}
  int initialize() {
    return 0;
  }
  ~qubx_idealize_mdl_context()
  {}
  
  qubx_idealizer_mdl* new_idealizer() {
    return new qubx_idealizer_mdl;
  }
};

// public interface:
extern "C" QUBFAST_API void* qubx_idealize_mdl_context_create()
{
  qubx_idealize_mdl_context * ctx = new qubx_idealize_mdl_context;
  if ( ctx->initialize() ) {
    delete ctx;
    return NULL;
  }
  return ctx;
}

extern "C" QUBFAST_API void qubx_idealize_mdl_context_free(void *ctx)
{
  delete (qubx_idealize_mdl_context *) ctx;
}

extern "C" QUBFAST_API void* qubx_idealizer_mdl_create(void *ctx)
{
  return ((qubx_idealize_mdl_context *) ctx)->new_idealizer();
}

extern "C" QUBFAST_API void qubx_idealizer_mdl_free(void *idlz)
{
  delete (qubx_idealizer_mdl *) idlz;
}
  
extern "C" QUBFAST_API int qubx_idealizer_mdl_idealize(void *idlz, float *data, int N, int Lmin, int split3, int combineClose, float combineDelta)
{
  return ((qubx_idealizer_mdl *) idlz)->idealize(data, N, Lmin, split3, combineClose, combineDelta);
}

extern "C" QUBFAST_API void qubx_idealizer_mdl_interrupt(void *idlz)
{
  ((qubx_idealizer_mdl *) idlz)->interrupt();
}

extern "C" QUBFAST_API int qubx_idealizer_mdl_get_dwells(void *idlz, int *counts, float *amps, float *stds, int capacity)
{
  return ((qubx_idealizer_mdl *) idlz)->get_dwells(counts, amps, stds, capacity);
}

extern "C" QUBFAST_API int qubx_idealizer_mdl_get_dwells_fl(void *idlz, int first_index, int *ff, int *ll, float *amps, float *stds, int capacity)
{
  qubx_idealizer_mdl *idl = (qubx_idealizer_mdl *) idlz;
  int N = idl->get_dwells(ll, amps, stds, capacity);
  int f = first_index;
  for (int i=0; i<N; ++i) {
    ff[i] = f;
    f += ll[i];
    ll[i] = f - 1;
  }
  return N;
}

